#ifndef INITUTIL_H
#define INITUTIL_H

#include "../base/base.h"
#include "../base/gmmdesc.h"
#include "../base/commonutil.h"

#include <iostream>

namespace initutil
{

	/**
	 * uses uniformMeans, then runs substeps many steps of the classical Lloyds algorithm and finally uses means2GMM.
	 */
	GMMDesc uniformLloydsMeans2GMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, unsigned int substeps);
	
	/**
	 * uses uniformMeans, then runs substeps many steps of the classical Lloyds algorithm and finally uses means2GMM.
	 */
	GMMDesc gonzalezLloydsMeans2GMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, unsigned int substeps);
	
	/**
	 * uses adaptiveMeans, then runs substeps many steps of the classical Lloyds algorithm and finally uses means2GMM.
	 */
	GMMDesc adaptiveLloydsMeans2GMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, unsigned int substeps);
	
	/**
	 * 
	 */
	GMMDesc alternatingAdaptiveMeanAndMeanToGMMThenLloydsForGMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, fp_type factor, bool spherical, unsigned int substeps, bool sphericalLloyds);
	
	/**
	 * 
	 */
	GMMDesc alternatingAdaptiveMeanAndMeanToGMMThenLloyds(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, fp_type factor, bool spherical, unsigned int substeps);
	
	
	/**
	 * 
	 */
	GMMDesc gonzalezForGMMThenLloydsForGMM(const commonutil::DataSet& input, idx_type k, std::mt19937& gen, bool use2GMM, bool spherical, fp_type sampleFactor, unsigned int substeps, bool sphericalLloyds);

	
	/**
	 * chooses k means uniformly at random from the given input set.
	 */
	Matrix uniformMeans(commonutil::DataSet const& input, idx_type k,  std::mt19937& gen,  std::vector<idx_type> const& indices_of_points_to_be_used = std::vector<idx_type>());

	
	/**
	 * computes means according to initutil::uniformMeans, then uses these means to
	 * compute a GMM by using kmeans::meansToGMM.
	 */
	GMMDesc uniformMeansToGMM(commonutil::DataSet const& input, idx_type k,  std::mt19937& gen);
	
	/**
	 * computes means according to initutil::uniformMeans, then uses these means
	 * as centers of a GMM by using kmeans::wrapMeans. 
	 */
	GMMDesc wrappedUniformMeans(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, bool computeWeightAndCovar = 1);
	
	/**
	 * chooses k means according to kmeans++.
	 */
	Matrix adaptiveMeans(commonutil::DataSet const& input, idx_type k, std::mt19937& gen);

	
	/**
	 * computes means according to initutil::adaptiveMeans, then uses these means to
	 * compute a GMM by using kmeansutil::meansToGMM.
	 */
	GMMDesc adaptiveMeansToGMM(commonutil::DataSet const& dataset, idx_type k, std::mt19937& gen);
	
	/**
	 * computes means according to initutil::adaptiveMeans, then uses these means
	 * as centers of a GMM by using kmeansutil::wrapMeans.
	 */
	GMMDesc wrappedAdaptiveMeans(commonutil::DataSet const& dataset, idx_type k, std::mt19937& gen, bool computeWeightAndCovar = 1);

	/**
	 * initialization according to Dasgupta&Schulman but without pruning.
	 */
	GMMDesc uniformSphericalGMM(commonutil::DataSet const& dataset, idx_type k, std::mt19937& gen);
	
	/**
	 * initialization according to Dasgupta&Schulman.
	 * 
	 * @param oversamplingFactor determines how many components are sampled before pruning, i.e., oversamplingFactor*k*ln(oversamplingFactor*k)
	 */
	GMMDesc uniformSphericalGMMwithPruning(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, unsigned int oversamplingFactor = 2);

	/**
	 * Gonzales-like pruning according to Dasgupta&Schulman.
	 */
	GMMDesc prune(GMMDesc& desc, idx_type k, std::mt19937& gen);
  
	/**
	 * adaptive initialization similar to Dasgupta&Schulman.
	 */
	GMMDesc adaptiveSphericalGMM(commonutil::DataSet const& dataset, idx_type k, std::mt19937& gen);
	
	/**
	 * alternately performs two steps: Chooses a point wrt gmmutil::adaptiveDensities and
	 * calls kmeansutil::meansToGMM wrt the sampled point and the means of the current gmm.
	 */
	GMMDesc alternatingAdaptiveMeanAndMeanToGMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, fp_type factor, bool spherical);

	
	/**
	 * runs agglomerativeMeans and calls kmeansuitl::meansToGMM wrt the input set and the returned means.
	 */
	GMMDesc sampleAgglomerativeMeansToGMM(const commonutil::DataSet& input, const idx_type k, const fp_type sampleSizeFactor, const bool precompute, std::mt19937& gen);

	/**
	 * computes average linkage clustering. 
	 * TODO: only works with unweighted points
	 */
	Matrix agglomerativeMeans(commonutil::DataSet const&, idx_type, bool = false);
	
	/**
	 * implements Gonazalez algorithm.
	 */
	Matrix gonzalez(commonutil::DataSet const& input, idx_type k, std::mt19937& gen);
	
	/**
	 * an version of the Gonzalez algorithm that draws means wrt maximum likelihood and uses means2gmm to estimate the covariances.
	 */
	GMMDesc gonzalezForGMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, bool use2GMM, bool spherical, fp_type sampleFactor);
	
	/**
	 * an version of the Gonzalez algorithm that draws means wrt maximum likelihood and uses random covariances.
	 */
	GMMDesc gonzalezKwedlo(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, fp_type sampleFactor);
	
	/**
	 * runs Gonzalez algorithm and wraps the Means using kmeansutil::wrapMeans.
	 */
	GMMDesc wrappedGonzalez(commonutil::DataSet const& input, idx_type k, std::mt19937& gen, bool computeWeightAndCovar);
	
	/**
	 * runs Gonzalez algorithm and calls kmeansutil::meansToGMM.
	 */
	GMMDesc gonzalezToGMM(commonutil::DataSet const& input, idx_type k, std::mt19937& gen);
	
	
	/**
	 * throws an exception if k==0 || d==0 || n==0.
	 */
	void check(idx_type k, idx_type d, idx_type n);
	
}

#endif